/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast                                     */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol;

import jarol.exceptions.AwaitDuringSignalingException;
import jarol.exceptions.SignalDuringSignalingException;
import jarol.exceptions.SignalNotAwaitedException;


/**
 * Abstract class for controlling a vehicle that is connected to a vehicle and a control terminal (terminal unit + navigation
 *  unit).
 *  
 * It provides three synchronization points in form of signals. One to activate the controller, when the sensor has finished its
 * execution. One to activate the terminal interface, when the controller has finished passing the display data to the 
 * terminal. The last one to activate the actuator interface, when the controller has finished computing the actuation data 
 * forwarded it to the actuator interface. 
 * <br>
 * It should be noted that the controller doesn't wait for navigation data from
 * the terminal. An additional Signal is needed for this kind of behavior, whereas the method <code>awaitNavigation()</code>
 * is provided in the interface and called in the <code>controlLoop()</code>. Hence an additional Signal and the implementation 
 * of <code>awaitNavigation()</code> would make the controller wait for the navigation data.
 * 
 * <br>
 * 
 * 
 * @author Bernhard Kast
 * @version 0.1
 */
public abstract class JarolCore implements JarolCoreInterface
{

	// Signals
	/**
	 * The signal that awaits on the sensor interface.
	 */
	protected Signal signalSensorController;
	/**
	 * The signal that signals the actuator interface.
	 */
	protected Signal signalControllerActuator;
	/**
	 * The signal that signals the terminal interface.
	 */
	protected Signal signalControllerTerminal;


	
	// Ports
	/**
	 * The port is used to transfer the sensor data from the sensor interface to the controller interface (Jarol Core).
	 */
	protected Port portSensorController; 
	/**
	 * The port is used to transfer the sensor data from the controller interface (Jarol Core) to the terminal interface.
	 */
	protected Port portControllerTerminal; 
	/**
	 * The port is used to transfer the navigation data from the navigation interface to the controller interface (Jarol Core).
	 */
	protected Port portNavigatorController; 
	/**
	 * The port is used to transfer the sensor data from the controller interface (Jarol Core) to the acutator interface.
	 */
	protected Port portControllerActuator; 


	/**
	 * The condition that indicates if the <code>controlLoop()</code> should be looping.
	 */
	protected boolean condition = true;

	

	/** 
	 * Loops while the condition is true and calls all methods needed for the control in a appropriate order.
	 */
	public void controlLoop()
	{
		MessageInterface[] sensorMsg, naviMsg, acutatorMsg;
		while(condition)
		{
			awaitSensor();
			sensorMsg = readFromSensor();
			
			forwardToTerminal(sensorMsg);
			signalTerminal();
			
			forwardToNavigation(sensorMsg);
			signalNavigation();
			
			awaitNavigation();
			naviMsg = readFromNavigation();
			
			acutatorMsg = computeActuationData(sensorMsg, naviMsg);
			forwardToActuator(acutatorMsg);
			signalActuator();
		}

	}

	public abstract void awaitNavigation();

	public abstract MessageInterface[] computeActuationData(MessageInterface[] sensorData,
			MessageInterface[] navigationData);
	
	public abstract void forwardToActuator(MessageInterface[] actuatorData);


	public abstract void forwardToNavigation(MessageInterface[] sensorData);

	public abstract void forwardToTerminal(MessageInterface[] dataBundle);

	public abstract MessageInterface[] readFromNavigation();

	public abstract MessageInterface[] readFromSensor();


	public abstract void signalNavigation();


	/** 
	 * Awaits a signal on the signalSensorController to continue in the <code>controlLoop()</code>.
	 */
	public void awaitSensor()
	{
		try
		{
			signalSensorController.await();
			
		} catch (AwaitDuringSignalingException e)
		{
			System.err.println("AwaitDuringSignalingException in awaitSensor().");
			e.printStackTrace();
		}
	}
	
	/**
	 * Sends a signal to the signalControllerActuator to initiate the continuation of the actuator interface.
	 */
	public void signalActuator()
	{
		try
		{
			signalControllerActuator.signal();
		} catch (SignalDuringSignalingException e)
		{
			System.err.println("SignalDuringSignalingException in signalActuator().");
			e.printStackTrace();
		} catch (SignalNotAwaitedException e)
		{
			System.err.println("SignalNotAwaitedException in signalActuator().");
			e.printStackTrace();
		}
	}
	
	/**
	 * Sends a signal to the signalControllerTerminal to initiate the continuation of the terminal interface.
	 */
	public void signalTerminal()
	{
		try
		{
			signalControllerTerminal.signal();
			
		} catch (SignalDuringSignalingException e)
		{
			System.err.println("SignalDuringSignalingException in signalTerminal().");
			e.printStackTrace();
		} catch (SignalNotAwaitedException e)
		{
			System.err.println("SignalNotAwaitedException in signalTerminal().");
			e.printStackTrace();
		}
	}


}
