/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast                                     */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;




/**
 * The JarolCoreInterface should be implemented by the class that performs
 * the control operations.
 * 
 * <br>Depending on the architecture some methods need not to be implemented, e.g., 
 * <code>signalNavigation()</code> is only needed if your design requires
 * that the navigation unit is activated at a deterministic point.
 * 
 * <br><br>NOTE: The internal Jarol components are called interfaces, e.g., sensor interface,
 * whereas the outside components are called units, e.g., navigation unit.
 * 
 * @author Bernhard Kast
 * @version 0.1
 */



public interface JarolCoreInterface {
	
	/**
	 * Calculates the information (e.g., motor signals, thruster, fin positions, ...) 
	 * and returns an array of messages containing this information.
	 * 
	 * @param sensorData data provided by the sensor interface
	 * @param navigationData data provided by the navigation interface
	 * @return computed actuation data - acutationData
	 */

	public MessageInterface[] computeActuationData(MessageInterface[] sensorData, MessageInterface[] navigationData);
	

	/**
	 * Should be called in the <code>run()</code> method and contain a
	 * loop that calls all the necessary methods of JarolCoreInterface.
	 *
	 */
	public void controlLoop();
	
	/**
	 * Waits for a signal from the sensor interface that indicates that the control
	 * flow can continue.
	 *
	 */
	public void awaitSensor();
	
	
	/**
	 * Reads data from the sensor interface.
	 * @return sensor data
	 */
	public MessageInterface[] readFromSensor();
	
	/**
	 * Signals the actuator interface to continue.
	 *
	 */
	
	public void signalActuator();
	
	/**
	 * Fordwards the actuation data to the actuator interface.
	 *
	 *@param actuatorData actuation data
	 */
	public void forwardToActuator(MessageInterface[] actuatorData);
	
	
	/** 
	 * Signals the terminal interface to continue.
	 *
	 */
	
	public void signalTerminal();
	
	
	/**
	 * Forwarding a dataBundle to the terminal, e.g. sensorData,
	 * actuatorData, etc. depending on the architecture and the 
	 * information that should be displayed/used at the terminal.
	 *
	 */
	
	public void forwardToTerminal(MessageInterface[] dataBundle);
	
	
	
	/**
	 * Forwards sensorData to the navigation interface. NOTE: The sensorData may
	 * be null if the navigation unit is non-autonomous, e.g., steering with
	 * a joystick.
	 * 
	 * @param sensorData 
	 */
	public void forwardToNavigation(MessageInterface[] sensorData);
	
	/**
	 * Waits for a signal from the navigation interface that indicates that the control
	 * flow can continue.
	 */
	public void awaitNavigation();
	
	/**
	 * Reads navigation data from the navigation interface.
	 * @return navigationData
	 */
	public MessageInterface[] readFromNavigation();
	
	/**
	 * Signals the navigation interface.
	 *
	 */
	public void signalNavigation();
	

}
