/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast                                     */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;

import jarol.messages.Link;
import jarol.messages.MessageFactory;
import jarol.messages.UDPLink;

import java.io.IOException;
import java.net.SocketException;


/**
 * Abstract class for controlling a vehicle that is connected to a vehicle and a control terminal (terminal unit + navigation
 *  unit) via UDPLinks.
 * 
 * It provides three synchronization points in form of signals. One to activate the controller, when the sensor has finished
 *  its execution. One to activate the terminal interface, when the controller has finished passing the display data to the 
 * terminal. The last one to activate the actuator interface, when the controller has finished computing the actuation data 
 * forwarded it to the actuator interface. 
 * <br>
 * It should be noted that the controller doesn't wait for navigation data from
 * the terminal. An additional Signal is needed for this kind of behavior, whereas the method <code>awaitNavigation()</code>
 * is provided in the interface and called in the <code>controlLoop()</code>. Hence an additional Signal and the implementation 
 * of <code>awaitNavigation()</code> would make the controller wait for the navigation data.
 * 
 * <br>
 * 
 * 
 * @author Bernhard Kast
 * @version 0.1
 */

public abstract class JarolCoreWithLink extends JarolCore
{

	
	/**
	 * The Link that provides an input and output port to the vehicle. Used for transfering the sensor and actuation data
	 * into and out of Jarol.
	 */
	protected Link vehicleLink;
	/**
	 * The Link that provides an input and output port to the terminal. NOTE: This assumes that the terminal also
	 * provides navigation data.
	 */
	protected Link terminalLink;
	
		
	
	/**
	 * Creates an UDPLink on the localPort to a given host on its remotePort.  
	 * 
	 * @param messageFactory a concrete MessageFactory
	 * @param host the hostname
	 * @param localPort the local port
	 * @param remotePort the remote port
	 * @param receiveBufferSize the size for the receiving buffer
	 * @param sendBufferSize the size for the sending buffer
	 * @return link the connected link, null if an error occured.
	 */
	protected UDPLink setupUDPLink(MessageFactory messageFactory, String host, int localPort, int remotePort, int receiveBufferSize, int sendBufferSize)
	{
		UDPLink link = null;
		
		try
		{
			link = new UDPLink(messageFactory, host, localPort, remotePort,receiveBufferSize,sendBufferSize);
			link.connect();

		} catch (SocketException e)
		{
			System.err.println("SocketException in initiateConnection()");
			e.printStackTrace();
		} catch (IOException e)
		{
			System.err.println("IOException in initiateConnection()");
			e.printStackTrace();
		}
		if (link == null)
		{
			System.err.println("establishing connection to host " + host + " failed.");
		}
		else
		{
			System.out.println("Connection to " + host + " established.");
		}
		return link;
	}

	
	
	


}
