/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast, Christoph Kirsch                   */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;




/** Port is a data transfer point that provides concurrent access for a single-reader and 
 * a single-writer. There is no support for multiple readers or writers. All items that are put and taken from the buffer
*  are cloned, also all items must implement the <code>MessageInterface</code> to provide cloning.
* 
* @author Bernhard Kast, Christoph Kirsch
* @version 0.1
*/

public class Port 
{
	/**
	 * buffer size
	 */
	private final int bufferSize; //final: set once
	/**
	 * content of the buffer
	 */
	private MessageInterface bufferContent[]; //buffer
	/**
	 * indicates the actual position of the writer in the circular buffer, but in
	 * 2 steps, hence the index is writeCounter/2.
	 */
	private int writeCounter = 0;//index for the writer
	/**
	 * indicates the actual position of the reader in the circular buffer, but in
	 * 2 steps, hence the index is readCounter/2.
	 */
	private int readCounter = 0;//index for the reader


	
	/**
	 * Constructs a port with the standard buffer size.
	 */
	public Port()
	{
		this(5);//default value
	}
	
	/**
	 * Creates a port with a specified buffer size.
	 * @param size indicates the size of the buffer
	*/
	
	public Port(int size)
	{
		if (size > 0)
		{
			bufferSize = size + 1;// plus one, because one field of the buffer is used as "delimiter"
			bufferContent = new MessageInterface[bufferSize];
			for(int i=0;i<bufferSize;i++)
				bufferContent[i] = null; //init buffer with null
		}
		else
		{
			throw new IllegalArgumentException("Buffer size must be greater than 0"); 
		}
	}
	
	
		
	/**
	 * Reads the oldest element from the buffer without blocking. If the buffer is empty null is returned.
	 * @return oldest element of the buffer
	 */

	public MessageInterface read()
	{
		return readSimple();
	}

	private MessageInterface readBuffer(int index)
	{
		return bufferContent[index];
	}

	/**
	 * Writes an item into the buffer. When the buffer is already full, false is returned and
	 * the buffer is unchanged.
	 * @param item - The element to be added into the buffer.
	 * @return true - if the item was successfully copied into the buffer, false - if the buffer is full
	 */

	public boolean write(MessageInterface item)
	{
		return writeSimple(item);
	}
	
	private boolean writeSimple(MessageInterface item)
	{
		//if ((readCounter == 0 && writeCounter + 1 == bufferSize) || writeCounter + 1 == readCounter)
			//return false;
		if (readCounter == 0)//reader is at the beginning
		{
			if (writeCounter + 1 == bufferSize) //writer is at the end
				return false; //buffer full
		}
		else
		{
			if (writeCounter + 1 == readCounter) //can only be true, when (writeCounter + 1 == bufferSize)
				return false; //buffer full
		}
				
		writeBuffer(writeCounter, item);
		if (writeCounter + 1 == bufferSize)
			writeCounter = 0;
		else
			writeCounter++;
		return true;
			
		
	}
	
	private MessageInterface readSimple()
	{
		if (readCounter == writeCounter)
			return null;
		MessageInterface item = readBuffer(readCounter);
		if (readCounter + 1 == bufferSize)
			readCounter = 0;
		else
			readCounter++;
		return item;
	}


	private void writeBuffer(int index, MessageInterface item)
	{
		try
		{
			bufferContent[index] = (MessageInterface)item.clone();
		} catch (CloneNotSupportedException e)
		{
			System.out.println("CloneNotSupportedException in writebuffer() - Port.");
			e.printStackTrace();
		}
	}

	/** 
	 * Returns the buffer size.
	 * @return buffer size
	 */
		
	public int getBufferSize()
	{
		return bufferSize;
	}
	


	
}