/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast, Christoph Kirsch                   */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;




/** PortEnhanced is a data transfer point that provides concurrent access for a single-reader and 
 * a single-writer with enhanced semantics that will never block the writer. 
 * The internal buffer is a pushing buffer, thus the oldest value is
 * replaced with the newest value iff the buffer is full.
 *  There is no support for multiple readers or writers. All items that are put and taken from the buffer
*  are cloned, also all items must implement the <code>MessageInterface</code> to provide cloning.
* 
* @author Bernhard Kast, Christoph Kirsch
* @version 0.1
*/

public class PortEnhanced 
{
	/**
	 * buffer size
	 */
	private final int bufferSize; //final: set once
	/**
	 * content of the buffer
	 */
	private MessageInterface bufferContent[]; //buffer
	/**
	 * indicates the actual position of the writer in the circular buffer, but in
	 * 2 steps, hence the index is writeCounter/2.
	 */
	private int writeIndex = 0;//index for the writer
	/**
	 * indicates the actual position of the reader in the circular buffer, but in
	 * 2 steps, hence the index is readCounter/2.
	 */
	private int readIndex = 0;//index for the reader

	private int writeLevel;
	private int readLevel;
	

	
	/**
	 * Constructs an enhanced port with the standard buffer size.
	 */
	public PortEnhanced()
	{
		this(5);//default value
	}
	
	/**
	 * Creates an enhanced port with a specified buffer size.
	 * @param size indicates the size of the buffer
	*/
	
	public PortEnhanced(int size)
	{
		if ((size > 0) && (size <= Integer.MAX_VALUE/2))
		{
			bufferSize = size;
			bufferContent = new MessageInterface[bufferSize];
			for(int i=0;i<bufferSize;i++)
				bufferContent[i] = null; //init buffer with null
		}
		else
		{
			throw new IllegalArgumentException("Buffer size must be greater than 0 and smaller than Integer.MAX_VALUE / 2."); 
		}

		writeLevel = Integer.MIN_VALUE;
		readLevel = 0;
	
	}
	
	
	private MessageInterface readBuffer(int index)
	{
		return bufferContent[index];
	}


	private void writeBuffer(int index, MessageInterface item)
	{
		try
		{
			bufferContent[index] = (MessageInterface)item.clone();
		} catch (CloneNotSupportedException e)
		{
			System.out.println("CloneNotSupportedException in writebuffer() - PortEnhanced.");
			e.printStackTrace();
		}
	}

	/** 
	 * Returns the buffer size.
	 * @return buffer size
	 */
		
	public int getBufferSize()
	{
		return bufferSize;
	}
	
	/**
	 * Reads the oldest element from the buffer without blocking. If the buffer is empty null is returned.
	 * @return oldest element of the buffer
	 */

	
	public MessageInterface read()
	{	
		while(true)
		{
			int savedWriteIndex = writeIndex;
			int savedWriteLevel = writeLevel;
			
			if ((savedWriteIndex & 1) == 0) //  even?
			{
                // writeIndex and writeLevel are consistent ?
				if ((savedWriteIndex == writeIndex) && (savedWriteLevel == writeLevel)) 
				{
					if (savedWriteLevel == readLevel)
					{
						if (readIndex == savedWriteIndex) //same slot & same wrap level
							return null; //buffer empty
					}
					else
						//ASSUMPTION: Integer.MIN_VALUE == Integer.MAX_VALUE + 1 
						if ((savedWriteLevel != readLevel + 1) || (savedWriteIndex > readIndex)) //push reader to the oldest element
						{
							//note these lines, don't need a successful read! Hence no rollback is necessary in a failed read attempt
							readIndex = savedWriteIndex;
                            //ASSUMPTION: Integer.MAX_VALUE == Integer.MIN_VALUE - 1
							readLevel = savedWriteLevel - 1; //oldest element
						}
					
					MessageInterface item = readBuffer(readIndex >> 1);//read
					
					if ((savedWriteIndex == writeIndex) && (savedWriteLevel == writeLevel)) //	no write occured meanwhile
					{
						readIndex = readIndex + 2; 
						if (readIndex == bufferSize*2)
						{
							readIndex = 0;
                            //ASSUMPTION: Integer.MIN_VALUE == Integer.MAX_VALUE + 1 
							readLevel = readLevel + 1;
						}
						return item;
					}
					
				}
			}
		}
	}



	
	
	/**
	 * Writes an item into the buffer. When the buffer is already full, the oldest element
	 * of the buffer is overwritten with item.
	 * @param item - The element to be added into the buffer.
	 */

	public void write(MessageInterface item)
	{
		writeIndex = writeIndex + 1; //writeIndex is odd; can't be buffersize*2 because it is odd.
		
		writeBuffer(writeIndex >> 1, item); //write
		
        //check if a wraparound will happen (can't do after writeCounter update, because it would "unprotected"(odd/even))
		if (writeIndex + 1 == bufferSize*2)  
		{
			writeLevel = writeLevel + 1;
			writeIndex = 0;//writeIndex is now even
		}
		else
			writeIndex = writeIndex + 1; //writeIndex is now even
	}
	



}