/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast                                     */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;

import jarol.exceptions.AwaitDuringSignalingException;
import jarol.exceptions.SignalDuringSignalingException;
import jarol.exceptions.SignalNotAwaitedException;



/**
 * Signal is a synchronization point, the <code>await()</code> call causes the 
 * callers (e.g., thread, process, exotask,...) to be suspended until a 
 * <code>signal()</code> call is performed by an arbitary task.
 * 
 * <br><br>
 * A SignalException is thrown in the following cases:<br>
 * (1) <code>signal()</code> is invoked and there is no thread awaiting. - SignalNotAwaitedException<br>
 * (2) <code>signal()</code> is invoked while signaling is in progress (during the wake up phase of awaiting threads). 
 * - SignalDuringSignalingException<br>
 * (3) <code>await()</code> is invoked while signaling is in progress. - AwaitDuringSignalingException<br>
 * (4) the awaiting thread is interrupted (<code>interrupt()</code> called).<br>
 * 
 * @author Bernhard Kast
 * @version 0.1
 *
 */

public class Signal 
{
	private String name;
	private boolean condition = false;
	/**
	 * Indicates if there are one or more watching tasks.
	 */
	private int watching = 0; 
	private boolean debug = false;
	
	/**
	 * Constructs a signal.
	 */
	public Signal()
	{

	}

	/**
	 * Creates a signal with a specified name and the debug flag. This allows the developer 
	 * to watch the signaling sequence on the standard input/output.<br>
	 * The outputs are as follows:<br>
	 * for a <code>signal()</code> ("SIGNAL " + name)<br>
	 * for a <code>await()</code> ("AWAIT " + name)<br>
	 * for a received <code>signal()</code> after an <code>await()</code> ("AWAIT " + name + " leaving!")<br>
	 * 
	 * @param name - The name of the Clock.
	 */
	public Signal(String name, boolean debug)
	{
		this.name = name;
		this.debug = debug;
	}
	
	/**
	 * All callers of <code>await()</code> tasks are resumed. A SignalException is thrown, if
	 * a signal is invoked without a caller awaiting or signal occured while signaling
	 * is in progress.
	 * @throws SignalDuringSignalingException 
	 * @throws SignalNotAwaitedException 
	 */
	
	public synchronized void signal() throws SignalDuringSignalingException, SignalNotAwaitedException
	{
		if (debug) 
			System.out.println("SIGNAL " + name);
		if (condition)
			throw new SignalDuringSignalingException();			
		if (watching < 1)
			throw new SignalNotAwaitedException();
		condition = true;
		notifyAll();
	}
	
	/**
	 * Causes the caller to be suspended until <code>signal()</code> is called.
	 * A SignalException is thrown, if an <code>await()</code> is invoked without while signaling
	 * is in progress or a InterruptedException is caught.
	 * @throws AwaitDuringSignalingException 
	 *
	 */

	public synchronized void await() throws AwaitDuringSignalingException
	{
		if (debug) 
			System.out.println("AWAIT " + name);
		if (condition)
			throw new AwaitDuringSignalingException();
		watching++;
		
		try
		{
			while (!condition) //in order to avoid "wrong" notifications/wake-ups
				wait();
			if (debug)
				System.out.println("AWAIT " + name + " leaving!");
		} 
		catch (InterruptedException e)
		{
			throw new RuntimeException("awaiting interrupted.");
		}
		finally //executed in anyway
		{
			watching--;
			if (watching < 1)
				condition = false;
		}
	}
}
