/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Bernhard Kast                                     */
/* 																			 */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/

package jarol;

import jarol.exceptions.AwaitDuringSignalingException;
import jarol.exceptions.SignalDuringSignalingException;
import jarol.exceptions.SignalNotAwaitedException;




/**
 * TimeTrigger is a periodic signal that provides timing.
 * 
 * @author Bernhard Kast
 * @version 0.1
 *
 */

public class TimeTrigger extends Thread implements TimeTriggerInterface {
	
	private Signal signal;
	private int period;
	private boolean keepRunning = true;

	
	/**
	 * Allocates a new TimeTrigger with the given period. 
	 * @param period - The period in milliseconds.
	 */
	public TimeTrigger(int period)
	{
		this(period, false);
	}
	
	/**
	 * Allocates a new TimeTrigger with the given period and the debug flag.
	 * @param period - The period in milliseconds.
	 * @param debug - Print debug information.
	 */
	public TimeTrigger(int period,boolean debug)
	{
		signal = new Signal("TimeTrigger", debug);
		this.period = period;
	}
	
	/**
	 * Causes the calling task to be suspended until the period is finished.
	 * @throws AwaitDuringSignalingException 
	 *
	 */
	public void await() throws AwaitDuringSignalingException
	{
		signal.await();
	}
	
	/**
	 * Sets the period in milliseconds.
	 * @param period - The period in milliseconds.
	 */
	public void setPeriod(int period)
	{
		this.period = period;
	}
	/**
	 * Returns the current period.
	 * @return - The current period in milliseconds.
	 */
	public int getPeriod()
	{
		return period;
	}
	
	/**
	 * Performs the sleeping and waking up.
	 */
	
	public void run()
	{
		System.out.println("run() in TimeTrigger");
		//Thread.yield();//try: make sure await() is executed before sleep can finish, when starting the whole thing
		while(keepRunning)
		{
			try 
			{
				sleep(period);
				if (keepRunning) //check if the thread was stopped during the sleep
					signal.signal(); //wakeup all task that watch this clock 
			} 
			catch (InterruptedException e) 
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
			} catch (SignalDuringSignalingException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
			} catch (SignalNotAwaitedException e)
			{
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}
		
	}
	
	/**
	 * Stops the execution of the thread.
	 *
	 */

	public void stopExecution()
	{
		System.out.println("stopExecution() in TimeTrigger");
		keepRunning = false;
	}


}
