/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;

/**
 * Buffer class for serialization.
 * 
 * <p>This class defines the basic methods for interfacing with a buffer encoded
 * in big endian or small endian byte order.</p>
 * 
 * <p>It is essentially an extension wrapper for <i>java.nio.ByteBuffer</i>
 * which cannot be subclassed due to the particular mechanism of creating 
 * <i>ByteBuffer</i> objects (ie factory methods), 
*  thus the wrapper functionality (instead of subclassing)</p>.
 * 
 * <p>Note: The <i>buf</i> field is protected for fast access through eventual
 * subclasses.</p>
 *
 * @author Eduardo Marques
 * @version 0.1
 */

public class Buffer {

  /**
   * Charset decoder for reading string values.
   * Currently it is fixed for the 'ISO-8859-1' charset.
   */
  private final static CharsetDecoder csDec
     = Charset.forName("ISO-8859-1").newDecoder();

  /**
   * Charset encoder for writing string values.
   * Currently it is fixed for 'ISO-8859-1' charser.
   */
  private final static CharsetEncoder csEnc 
    = Charset.forName("ISO-8859-1").newEncoder();

  /**
   * Internal ByteBuffer instance, can be accessed by subclasses.
   */
  protected ByteBuffer buf;

  /**
   * Constructor from a data byte array. 
   * Little Endian byte order will be used by default.
   * @param data the source data
   */
  public Buffer(byte[] data){
    this(data,0,data.length,ByteOrder.LITTLE_ENDIAN);
  }
  /**
   * Constructor from a data byte array, 
   * plus specified offset and usable length.
   * Little Endian byte order will be used by default.
   * @param data the source data
   * @param off the source data offset
   * @param len the source data length
   */
  public Buffer(byte[] data, int off, int len){
    this(data,off,len,ByteOrder.LITTLE_ENDIAN);
  }

  /**
   * Constructor from a data byte array and specified byte order.
   * @param data the source data
   * @param order byte order to use
   */
  public Buffer(byte[] data,ByteOrder order){
    this(data,0, data.length,order);
  }

  /**
   * Constructor from a data byte array, 
   * plus specified offset, usable length and byte order to use.
   * @param data the source data
   * @param off the source data offset
   * @param len the source data length
   * @param order the byte order to use
   */
  public Buffer(byte[] data, int off, int len, ByteOrder order){
    reset(data, off, len, order);
  }
  /**
   * Reset buffer. 
   * @param data the source data
   * @param off the source data offset
   * @param len the source data length
   * @param order the byte order to use
   */
  public void reset(byte[] data, int off, int len, ByteOrder order){
    buf = ByteBuffer.wrap(data, off, len);
    buf.order(order);
  }
  /**
   * Get data array.
   * @return the byte array 
   */
  public byte[] getData(){
    return buf.array();
  } 
  /**
   * Rewind buffer to starting position.
   */
  public void rewind(){ buf.rewind(); }

  /**
   * Change byte order in use.
   * @param order the byte order to use (little endian or big endian)
   */
  public void setByteOrder(ByteOrder order){
     buf.order(order);
  }
  /**
   * Get byte order in use (big endian or little endian).
   * @return byte order in use.
   **/
  public ByteOrder getByteOrder(){
    return buf.order();
  }
  /**
   * Get buffer position.
   */
  public final int position(){
    return buf.position();
  }

  /**
   * Read a signed 8 bit integer from the buffer.
   * @return 8-bit signed value read
   */
  public final byte readINT8() {
    return buf.get();
  }
  
  /**
   * Read an unsigned 8 bit integer from the buffer.
   * @return 8-bit unsigned value read
   */
  public final short readUINT8()  {
    byte b = buf.get();
    return (b >= 0 ? (short) b : (short) (Message.MAX_UINT8 + 1 + (short) b));
  }

  /**
   * Read a signed 16 bit integer from the buffer.
   * @return 16-bit signed value read
   */
  public final short readINT16() {
    return buf.getShort();
  }
  
  /**
   * Read an unsigned 16 bit integer from the buffer.
   * @return 16-bit unsigned value read
   */
  public final int readUINT16()  {
    return buf.getChar();
  }
  
  /**
   * Read a signed 32 bit integer from the buffer.
   * @return 32-bit signed value read
   */
  public final int readINT32() {
    return buf.getInt(); 
  }
  /**
   * Read an unsigned 32 bit integer from the buffer.
   * @return 32-bit unsigned value read
   */
  public final long readUINT32()  {
    int i32 = buf.getInt();
    return i32 >= 0 ? (long) i32 : (long) (Message.MAX_UINT32 + 1) + (long) i32;
  }
  /**
   * Read a signed 64 bit integer from the buffer.
   * @return 64-bit signed value read
   */
  public final long readINT64() {
    return buf.getLong(); 
  }
  /**
   * Read an unsigned 64 bit integer from the buffer (fake).
   * <br>This method does not actually return unsigned 64-bit values,
   * but 64-bit signed values (ie the Java 'long' scalar type).
   * <br>TODO: possibly handle this another way 
   * or implement it with big integer class.
   * @return 64-bit signed value read
   */
  public final long readUINT64()  {
    return buf.getLong();
  }
  /**
   * Read a 32-bit floating point value from the buffer.
   * @return 32-bit floating point value read
   */
  public final float readFP32()  {
    return buf.getFloat();
  }
  /**
   * Read a 64-bit floating point value
   * from the buffer.
   * @return 64-bit floating point value read
   */
  public final double readFP64()  {
    return buf.getDouble();
  }
  /**
   * Read raw data.
   * Raw bytes are prefixed by an unsigned 16-bit header stating 
   * the data length.
   * @return null for a 0 byte-buffer, a byte array otherwise
   */
  public final byte[] readRawData()  {
    int len = readUINT16();
    byte[] data = null;
    if(len > 0){
      buf.get(data = new byte[len], 0, len);
    }
    return data;
  }
  /**
   * Read plain text (string) data with optional length prefix specifier.
   * <br>If 'encodedLength' equals true then 
   * String contents are prefixed by an unsigned 16-bit header stating 
   * the string length. Otherwise the buffer will be read until
   * the null character is found.
   * <br>Note that current implementation will decode strings
   * using the ISO-8859-1 charset.
   * @param encodedLength encoded length flag
   * @return string value 
   */
  public final String readPlainText(boolean encodedLength)  {
    int length;
    byte[] bdata;
    if(encodedLength){
      length = readUINT16();
      if(length == 0) return "";
      bdata = new byte[length];
    }else{
      int i;
      int off=buf.position();
      int limit = buf.limit();
      for(i=off;i<limit && buf.get(i) != '\0';i++);
      length=i-off+1;
      System.out.println(off + " "+limit+ " "+length);
      bdata=new byte[length];
    }
    buf.get(bdata);
    int l = 0;
    for(l=0;l<length && bdata[l] != 0;l++);
    if(l == 0) return "";

    CharBuffer cb = CharBuffer.wrap(new char[l]);
    csDec.decode(ByteBuffer.wrap(bdata,0,l), cb, false);
    cb.rewind();
    return cb.toString();
  }
  /**
   * Write 8-bit signed integer value onto buffer.
   * @param i8 8-bit signed integer value
   */
  public final void writeINT8(byte i8) {
    buf.put((byte)(i8&0x00FF));
  }
  /**
   * Write 8-bit unsigned integer value onto buffer.
   * @param u8 8-bit unsigned integer value (8-bit truncation will be made)
   */
  public final void writeUINT8(short u8) {
    buf.put((byte)(u8 & 0x00FF));
  }
  
  /**
   * Write 16-bit signed integer value onto buffer.
   * @param i16 16-bit signed integer value
   */
  public final void writeINT16(short i16) {
    buf.putShort(i16);
  }
  /**
   * Write 16-bit unsigned integer value onto buffer.
   * @param u16 16-bit signed integer value (truncation to 16 bits will be made)
   */
  public final void writeUINT16(int u16)  {
    buf.putChar((char)u16);
  }
  /**
   * Write 32-bit signed integer value onto buffer.
   * @param i32 32-bit signed integer value
   */
  public final void writeINT32(int i32) {
    buf.putInt(i32);
  }
  /**
   * Write 32-bit unsigned integer value onto buffer.
   * @param u32 32-bit signed integer value (truncation to 32 bits will be made)
   */
  public final void writeUINT32(long u32)  {
    buf.putInt((int)u32 & 0xFFFFFFFF);
  }

  /**
   * Write 64-bit signed integer value onto buffer.
   * @param i64 64-bit signed integer value
   */
  public final void writeINT64(long i64) {
    buf.putLong(i64);
  }
  /**
   * Write 64-bit unsigned integer value onto buffer.
   * The parameter is actually a 64-bit signed integer,
   * which may be understood as an unsigned 64-bit value
   * at a receiver's end in C/C++ for example
   * @param u64 64-bit signed integer value
   */
  public final void writeUINT64(long u64)  {
    buf.putLong(u64);
  }
  /**
   * Write 32-bit floating point value onto buffer.
   * @param f32 32-bit floating point value
   */
  public final void writeFP32(float f32)  {
    buf.putFloat(f32);
  }
  /**
   * Write 64-bit floating point value onto buffer.
   * @param f64 64-bit floating point value
   */
  public final void writeFP64(double f64)  {
    buf.putDouble(f64);
  }
  /**
   * Write raw data with 2-byte length prefix specifier.
   * The outcome onto the buffer will be an unsigned 16-bit value encoding
   * the data length followed by the actual data.
   * @param data to write
   */
  public final void writeRawData(byte[] data) {
    writeUINT16(data == null ? 0 : (char) data.length);
    if(data != null) buf.put(data);
  }
  /**
   * Write plain text with optional 2-byte length prefix specifier.
   * <br>The outcome onto the buffer will be, if 'encodeLength' is true,
   * an unsigned 16-bit value encoding the string length 
   * followed by the actual text. Otherwise, only the text data
   * will be written followed by the null terminator.
   * <br>Note that current implementation will encode strings
   * using the ISO-8859-1 charset.
   * @param text text to write
   * @param encodeLength indicate that length should be encoded as a 2-byte prefix 
   */
  public final void writePlainText(String text, boolean encodeLength)  {
    int length = text == null ? 0 : text.length();
    if(encodeLength) writeUINT16(length);
    if(length > 0) 
      csEnc.encode(CharBuffer.wrap(text), buf, false );  
    if(!encodeLength) buf.put((byte)0);
  }

  /* Unit test
  public static void main(String[] args) throws Exception {
    ByteOrder[] botypes = new ByteOrder[] { ByteOrder.LITTLE_ENDIAN, ByteOrder.BIG_ENDIAN };
    
    for(int i=0;i<botypes.length;i++){
      ByteOrder bo = botypes[i];
      Buffer obuf = new Buffer(new byte[Message.MAX_UINT16*2],bo);
      Buffer ibuf = new Buffer(obuf.getData(),bo);
      System.out.println(bo);
      System.out.println(obuf.buf.order() + " " + ibuf.buf.order());
     short i16; 
      obuf.writeINT8(Message.MIN_INT8);   if(Message.MIN_INT8 != ibuf.readINT8()) throw new Exception("error");
      obuf.writeINT16(Message.MIN_INT16); if(Message.MIN_INT16 != (i16=ibuf.readINT16())) throw new Exception("error "+bo+ " "+i16+ " " + Message.MIN_INT16);
      obuf.writeINT32(Message.MIN_INT32); if(Message.MIN_INT32 != ibuf.readINT32()) throw new Exception("error");
      obuf.writeINT64(Message.MIN_INT64); if(Message.MIN_INT64 != ibuf.readINT64()) throw new Exception("error");
      
      obuf.writeINT8(Message.MAX_INT8);   if(Message.MAX_INT8 != ibuf.readINT8()) throw new Exception("error");
      obuf.writeINT16(Message.MAX_INT16); if(Message.MAX_INT16 != ibuf.readINT16()) throw new Exception("error");
      obuf.writeINT32(Message.MAX_INT32); if(Message.MAX_INT32 != ibuf.readINT32()) throw new Exception("error");
      obuf.writeINT64(Message.MAX_INT64); if(Message.MAX_INT64 != ibuf.readINT64()) throw new Exception("error");
      
      obuf.writeUINT8(Message.MIN_UINT8);   if(Message.MIN_UINT8 != ibuf.readUINT8()) throw new Exception("error");
      obuf.writeUINT16(Message.MIN_UINT16); if(Message.MIN_UINT16 != ibuf.readUINT16()) throw new Exception("error");
      obuf.writeUINT32(Message.MIN_UINT32); if(Message.MIN_UINT32 != ibuf.readUINT32()) throw new Exception("error");
      obuf.writeUINT64(Message.MIN_UINT64); if(Message.MIN_UINT64 != ibuf.readUINT64()) throw new Exception("error");
      
      obuf.writeUINT8(Message.MAX_UINT8);   if(Message.MAX_UINT8 != ibuf.readUINT8()) throw new Exception("error");
      obuf.writeUINT16(Message.MAX_UINT16); if(Message.MAX_UINT16 != ibuf.readUINT16()) throw new Exception("error");
      obuf.writeUINT32(Message.MAX_UINT32); if(Message.MAX_UINT32 != ibuf.readUINT32()) throw new Exception("error");
      obuf.writeUINT64(Message.MAX_UINT64); if(Message.MAX_UINT64 != ibuf.readUINT64()) throw new Exception("error");
  
      obuf.writeFP32(Message.MIN_FP32);     if(Message.MIN_FP32 != ibuf.readFP32()) throw new Exception("error");
      obuf.writeFP64(Message.MIN_FP64);     if(Message.MIN_FP64 != ibuf.readFP64()) throw new Exception("error");
      obuf.writeFP32(Message.MAX_FP32);     if(Message.MAX_FP32 != ibuf.readFP32()) throw new Exception("error");
      obuf.writeFP64(Message.MAX_FP64);     if(Message.MAX_FP64 != ibuf.readFP64()) throw new Exception("error");
      
      obuf.writeRawData(null);      if(ibuf.readRawData() != null) throw new Exception("error");
      
      byte[] ba = new byte[Message.MAX_UINT16];
      for(int j=0;j<ba.length;++j){
        ba[j] = (byte) j;
      }
      obuf.writeRawData(ba);        byte[] ba2 = ibuf.readRawData(); 
      
      if(ba.length != ba2.length) throw new Exception("error");
      
      for(int j=0;j<ba.length;++j){
        if(ba[j] != ba2[j]) throw new Exception("Error " + j);
      }
      
      String abcde = "ABCDEFGHIJKLMNOPQRTUVXWYZabcdefghijklmnopqrstuvwxyz01234567890";
      String abcde2;
      obuf.writePlainText(abcde, true); if(!ibuf.readPlainText(true).equals(abcde)) throw new Exception("error");
      obuf.writePlainText(abcde, false); if(!(abcde2=ibuf.readPlainText(false)).equals(abcde)) throw new Exception("error "+abcde + " "+abcde2);
      
    }
  }*/
}
