/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;

import jarol.Port;
import jarol.Signal;
import jarol.exceptions.AwaitDuringSignalingException; 
import java.io.IOException;

/**
 * Jarol-wise, non-blocking, lock-free bidirectional message 
 * link abstract class. 
 * <p>
 * A link has the intent of providing 
 * calling tasks a Jarol-wise signal/port interface for operation,
 * with deterministic operation and non-blocking behavior.</p>
 * <p>
 * This base class is extended by classes that implement
 * actual network links in the Jarol system 
 * or custom ones defined by the user. Derived classes need only to provide
 * link access functions (start/stop/send/receive functions), the 
 * application interface logic is within the base link class.
 * <p>
 * The link can be used to send and receive message
 * using two instances of base.Port and an instance
 * of base.Signal:
 * <p>
 * 1) A <b>receive port</b> which callers can poll for received messages.
 * The receive port is updated asynchronously (as soon as messages arrive).
 * <p>
 * 2) A <b>send port</b> in which callers can place messages to be sent
 * <p>
 * 3) A <b>send signal</b> to actually send the  
 * messages in the send port over the link. To send messages, the link caller 
 * should place a number of messages in the send port then trigger the send
 * signal.
 * <p>
 * Internally <i>Link</i> works as follows:
 * <p>
 * 1) A "receiver thread" polls the communication link and when it gets
 * a new message it places it on the receive port.
 * <p> 
 * 2) A "sender thread" that waits on the send signal,
 * then  sends data from the send port over the communication link.
 * <p>
 *
 * Note that a link will self-disconnect if the physical connection
 * has a fatal error.
 *
 * @author Eduardo Marques
 * @version 0.1
 */
public abstract class Link {
  /**
   * Connected flag.
   */
  boolean connected;

  /**
   * Message factory.
   * 
   * It provides the interface needed to serialize/unserialize messages.
   */
   private MessageFactory factory;
   
   /**
    * Default buffer capacity for the receive and send ports
    */
   public static final int DEFAULT_PORT_BUFSZ = 16;
 
   /**
    * Maximum message payload
    */
   public static final int MAX_MESSAGE_PAYLOAD = 2048;

   /**
    * Receive port
    * 
    * Every message that is received is placed onto the receive port.
    * A receive port reference can be obtained with getRecvPort() and
    * is valid 
    */
   private final Port recvPort;

   /**
    * Send port
    *
    * Internal port used by the link to wake up the sending thread.
    * when there are new messages to send.
    */
   private final Port sendPort;
   
  /**
   * Send signal
   */
   private final Signal sendSignal;
   
  /**
   * Receiver thread
   */
   private ReceiverThread receiverThread;
   
  /**
   * Sender thread.
   */
   private SenderThread senderThread;
   
  /**
   * Constructor specifying message factory.
   *
   * Receive and send ports will be defined with buffer capacity
   * equal to Port.DEFAULT_PORT_CAPACITY .
   *
   * @param factory the message factory
   */
  public Link(MessageFactory factory){
    this(factory, DEFAULT_PORT_BUFSZ, DEFAULT_PORT_BUFSZ);
  }
  /**
   * Constructor specifying message factory and receive/send 
   * port capacity.
   * 
   * @param factory the message factory
   * @param recvPortBufSz the buffer size for the receive port
   * @param sendPortBufSz the buffer size for the send port
   */
  public Link(MessageFactory factory, int recvPortBufSz, int sendPortBufSz){
    this.connected = false;
    this.factory = factory;
    recvPort = new Port(recvPortBufSz);
    sendPort = new Port(sendPortBufSz);
    sendSignal = new Signal();
  }

  /**
   * Check if link is connected.
   * This procedure should be called to ensure link is connected 
   * and no connection errors occurred in the meanwhile.
   * @return true if the link is connected.
   */
  public final boolean connected(){
    return connected; 
  }
  /**
   * Get send port.
   * The send port will remain the same throughout the link's 
   * life (whenever it is connected or not).
   * @return send port
   */
  public final Port getSendPort(){
    return sendPort;
  }
  /**
   * Get send signal.
   * The send signal will remain the same throughout the link's 
   * life (whenever it is connected or not).
   * @return send signal
   */
  public final Signal getSendSignal(){
    return sendSignal;
  }
  /**
   * Get receive port.
   * The receive port will remain the same throughout the link's 
   * life (whenever it is connected or not).
   * @return receive port
   */
  public final Port getRecvPort(){
    return recvPort;
  }

  /**
   * Connect the message link. 
   * The request is ignored if the link is already in a connected state. 
   * Otherwise this will (1) setup the connection
   * and (2) spawn the receive and send threads
   *
   * @throws IOException if there a problem in connecting the link 
   */
  public final void connect() throws IOException {
    if(connected) return;
    
    // See (1) described above 
    start(); 

    connected = true;
    // See (3) described abobe 
    senderThread = new SenderThread(this);
    receiverThread = new ReceiverThread(this);

    senderThread.start();
    receiverThread.start();
  }
  /**
   * Disconnect the message link. 
   * The request is ignored if the link is already in a disconnected state. 
   * Otherwise  this will: 
   * 1) wait for any pending sending / receiving operations 
   * to complete 2) stop the receiver and sender threads, 
   * and 3) disconnect the physical connection.  
   * 
   * Note that the link may be reconnected and 
   * any references to the sender and receive ports will remain valid.
   *
   * Note also that it is possible to restart the link.
   */ 
  public final void disconnect(){
    if(!connected) return;
    connected = false;

    // See (1) and (2) described above
    synchronized(senderThread){
      if(senderThread.active){
        try { sendSignal.signal(); } 
        catch(Exception e){ }
        try { senderThread.wait(); }
        catch(InterruptedException e){} 
      }
    }
    synchronized(receiverThread){
      if(receiverThread.active){
        try { receiverThread.wait(); }
        catch(InterruptedException e){} 
      }
    }
    // See (3) described above.
    receiverThread = null; 
    senderThread = null; 

    // See (4) described above.
    try { 
      stop();
    }catch(IOException e){
      System.err.println("A link specific disconnect error occurred:" + e.toString());
    }
    
  } 
  /**
   * Procedure to disconnect the link when a serious error occurs.
   */
  private void disconnectOnError(){
    new ShutdownThread(this).start();
  }
  /**
   * Abstract method for starting the link.
   * 
   * Implementations should define this method to allocate, setup and enable 
   * the actual communication link.
   * 
   * @throws IOException if a port-specific error occurs (eg network related)
   */
  protected abstract void start() 
  throws IOException;

  /**
   * Abstract method for stopping the link.
   *
   * Implementations should define this method to close and destroy the actual
   * communication link.
   *
   * @throws IOException if a port-specific error occurs (eg network related)
   */
  protected abstract void stop() 
  throws IOException;
  
  /**
   * Abstract method used to send data.
   * 
   * @param data the message payload
   * @param off payload offset
   * @param len payload length
   * @throws IOException if a port-specific error occurs (eg network related)
   */
  protected abstract void send(byte[] data, int off, int len) 
  throws IOException;

  /**
   * Abstract method used to receive data.
   *
   * @param buf the receive buffer to use
   * @param off the receive buffer offset
   * @param len the receive buffer usable length
   * @param timeout the time to wait for a message
   * @return the length of the received message packet
   * or 0 in case timeout has been reached
   * @throws IOException if a port-specific error occurs (eg network related)
   */
  protected abstract int recv(byte[] buf, int off, int len, int timeout) 
  throws IOException;

  /**
   * Internal receiver thread.
   */
  private static class ReceiverThread extends Thread {
    /**
     * Message link associated the receiver thread.
     */
    final Link link;
    
    /**
     * Active flag.
     */
    boolean active;

    /**
     * Constructor.
     * @param link the message link associated to the receiver thread
     */
    ReceiverThread(Link link){
      this.link = link;
      this.active = false;
    }
    /**
     * run() implementation.
     * <p>
     * The cycle of the reader thread is to:
     * <p> 
     * 1) wait for a message over
     * <p>
     * the communications link and 2) place it on the receiver port.
     */ 
    public void run() {
      Port port = link.recvPort;
      MessageFactory factory = link.factory;
      byte[] buf = new byte[MAX_MESSAGE_PAYLOAD]; 
      synchronized(this){ active = true; }

      while(link.connected){
        try {
          // Poll for messages
          int len = link.recv(buf, 0, buf.length, 100);
          // If len equals 0, that means a timeout occurred
           
          int used_len = 0;

          while(used_len < len){
            // Unserialize message 
            Message msg = factory.unserialize(buf, 0, len);
          
            used_len += 
              factory.HEADER_SIZE + msg.serialSize() + factory.FOOTER_SIZE;

            // Place message on the receiver port
            if(port.write(msg) == false){
              System.err.println("Link receive buffer full!!");
            }
          }
        }catch(InvalidMessageException e){
          e.printStackTrace();
          System.err.println("Link receive error: invalid message: "+e.getMessage());
        }catch(IOException e){
          e.printStackTrace();
          System.err.println("Link receive error: "+e.getMessage() +" - disconnecting link");
          link.disconnectOnError();
        }
      }
      /**
       * Let stopper thread know we're done
       */
      synchronized(this){ notify(); active = false;}
    }
  }
  /**
   * Internal sender thread.
   */
  private static class SenderThread extends Thread {
    /**
     * Message link associated the receiver thread.
     */
    final Link link;
    
    /**
     * Active flag
     */
    boolean active;

    /**
     * Constructor
     * @param link the message link associated to the receiver thread
     */
    SenderThread(Link link){
      this.link = link;
    }
      
    /**
     * run() implementation.
     * <p>
     * The cycle of the sender thread consists in:
     * <p>
     * 1) wait for a message over the send port 
     * <p>
     * 2) transmit the message over the link
     */ 
    public void run() {
      Port port = link.sendPort;
      MessageFactory factory = link.factory;
      byte[] buf = new byte[64]; // initial size, reallocated if necessary
      synchronized(this){ active = true; }

      while(link.connected){
        try {
          // Await send signal
          link.sendSignal.await();    
	  
          // Check send port for messages
          Message msg;

          while((msg = (Message)port.read())!=null){
            // Convert message to serialization format 
            int slzLen = 
              link.factory.HEADER_SIZE 
            + msg.serialSize() 
            + link.factory.FOOTER_SIZE;

            if(buf.length < slzLen){
              // Reallocate buffer, current length is not
              // enough to hold message payload
              buf = new byte[slzLen];
            }
            factory.serialize(msg, buf, 0, slzLen);
          
            // Send the message in serialized format
            link.send(buf, 0, slzLen);
          }
        }
	catch(AwaitDuringSignalingException e){
          System.err.println("Link synchronization error: AwaitDuringSignalingException");
        }
	catch(InvalidMessageException e){
          e.printStackTrace();
          System.err.println("Link send error: invalid message: "+e.getMessage());
        }catch(IOException e){
          e.printStackTrace();
          System.err.println("Link send error: "+e.getMessage() +" - disconnecting link ...");
          link.disconnectOnError();
        }
      }
      // Let stopper thread know we're done
      synchronized(this){ notify(); active = false; }
    }
  }
  /**
   * Shutdown thread. The thread is spawned on auto-disconnect, typically
   * when a serious link error occurs.
   */
  private static class ShutdownThread extends Thread {
    final Link link;
    /**
     * Constructor
     */
    public ShutdownThread(Link link){
      this.link = link;
    }
    /**
     * Execution of shutdown thread.
     * The run() method invokes the disconnect() method for a link.
     */
    public void run(){
      try { link.disconnect(); } catch(Exception e){ }
    }
  }
}
  
