/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;
import jarol.MessageInterface;


/**
 * Base class for messages.
 * @author Eduardo Marques
 * @version 0.1
 */
public abstract class Message implements MessagePart, MessageInterface {
  /**
   * Min value for UINT8 type.
   */
  public static final short MIN_UINT8 = 0;
  /**
   * Min value for UINT16 type.
   */
  public static final int MIN_UINT16 = 0;
  /**
   * Min value for UINT32 type.
   */
  public static final int MIN_UINT32 = 0;
  /**
   * Min value for UINT64 type.
   */
  public static final long MIN_UINT64 = 0L;

  /**
   * Min value for INT8 type.
   */
  public static final byte MIN_INT8 = Byte.MIN_VALUE;
  /**
   * Min value for INT16 type.
   */
  public static final short MIN_INT16 = Short.MIN_VALUE;
  /**
   * Min value for INT32 type.
   */
  public static final int MIN_INT32 = Integer.MIN_VALUE;
  /**
   * Min value for INT64 type.
   */
  public static final long MIN_INT64 = Long.MIN_VALUE;
  
  /**
   * Max value for INT8 type.
   */
  public static final byte MAX_INT8 = Byte.MAX_VALUE;
  /**
   * Max value for INT16 type.
   */
  public static final short MAX_INT16 = Short.MAX_VALUE;
  /**
   * Max value for INT32 type.
   */
  public static final int MAX_INT32 = Integer.MAX_VALUE;
  /**
   * Max value for INT64 type.
   */
  public static final long MAX_INT64 = Long.MAX_VALUE;
  
  /**
   * Max value for UINT8 type.
   */
  public static final short MAX_UINT8 = 255;
  /**
   * Max value for UINT16 type.
   */
  public static final int MAX_UINT16 = 65535;
  /**
   * Max value for UINT32 type.
   */
  public static final long MAX_UINT32 = 4294967295L;
  /**
   * Max value for UINT64 type (not accurate)
   */
  public static final long MAX_UINT64 = Long.MAX_VALUE;
  /**
   * Min value for FP32 type.
   */
  public static final float MIN_FP32 = Float.MIN_VALUE;
  /**
   * Max value for FP32 type.
   */
  public static final float MAX_FP32 = Float.MIN_VALUE;
  /**
   * Min value for FP64 type.
   */
  public static final double MIN_FP64 = Double.MIN_VALUE;
  /**
   * Max value for FP64 type.
   */
  public static final double MAX_FP64 = Double.MAX_VALUE;

  /**
   * No argument constructor.
   */
  public Message(){ }

  /**
   * Get timestamp.
   * By default this implementation always returns 0.0
   * It must be overridden by message sets which use header timestamps.
   */
  public double getTimestamp(){
    return 0.0;
  }
  /**
   * Set timestamp for message.
   * By default the base implemenation does nothing, nor Message
   * declares any field to hold the timestamp.
   * It must be overridden by message sets which use header sequence ids.
   */
  public void setTimestamp(double timestamp){
    /* dummy */
  }
  /**
   * Get sequence id 
   * By default this implementation always returns 0. 
   * It must be overridden by message sets which use header timestamps.
   */
  public long getSequenceId(){
   return 0;
  }
  /**
   * Set sequence id for message.
   * By default the base implemenation does nothing nor Message
   * declares any field to hold the sequence id.
   * It must be overridden by message sets which use header sequence ids.
   */
  public void setSequenceId(long id){
    /* dummy */
  }

  /**
   * Abstract method to validate messages.
   * The validate() method throws InvalidMessageException when
   * the message is invalid.
   */
  public abstract void validate() throws InvalidMessageException;
   
  /**
   * Get serialization id for message.
   * @return serialization id for message 
   */
  public abstract int serialId();
  /**
   * Get serialization size for message.
   * @return serialization size for message 
   */
  public abstract int serialSize();
 
  /**
   * Serialize message.
   * @param buffer Output buffer
   * @throws InvalidMessageException if the message is not valid
   */
  public abstract void serialize(Buffer buffer) throws InvalidMessageException;
  /**
   * Unserialize message.
   * @param buffer Input buffer
   * @throws InvalidMessageException if the unserialized message is not valid
   */
  public abstract void unserialize(Buffer buffer) throws InvalidMessageException;
  /**
   * deepClone() implementation. 
   *
   * This is a final method. Sub-classes should override clone() to 
   * provide deep-copy semantics if they contain 'rawdata' fields.
   *
   * @throws RuntimeException if an exception such as CloneNotSupportedException
   * was raised
   * @return deep clone of the message
   */
  public final Object deepClone() {
    try {
      return clone();
    }catch(Exception e){
      throw new RuntimeException(e);
    }
  }
  
  /**
   * clone() implementation with deep copy semantics.
   *
   * In order to provide compatibility with messages in the Jarol
   * system it should be overriden by message classes
   * with mutable fields.  Otherwise the shallow-copy
   * semantics of Object.clone() are sufficient.
   * @return deep clone of Message object
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone();
  }
  
  public static final boolean MSG_DEBUG;
  static {
    MSG_DEBUG=System.getProperty("MSG_DEBUG") != null;
  }
}
