/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;

/**
 * Message factory interface.
 *
 * <p>Each application message should define its own.
 * It defines factory methods to instantiate messages, message headers 
 * and footer.</p>
 *
 * @author Eduardo Marques
 * @version 0.1
 */
public abstract class MessageFactory {
  /**
   * Header size for messages of this factory
   * (0 if a message factory uses no headers)
   */
  public final int HEADER_SIZE;
  /**
   * Footer size for messages of this factory
   * (0 if a message factory uses no footers)
   */
  public final int FOOTER_SIZE;
  
  /**
   * Byte order used by this factory.
   */
  public final java.nio.ByteOrder BYTE_ORDER;
   
  /**
   * Checksum procedure 
   * If null, no checksum is performed and considered for message footers.
   */
  public final SerializationHandle slzHandle;

  /**
   * Constructor  
   * @param headerSize header size
   * @param footerSize footer size
   * @param byteOrder byte order to use (BIG_ENDIAN or LITTLE_ENDIAN)
   * @param slzHandle SerializationHandle instance
   */
  protected MessageFactory
  (int headerSize, 
   int footerSize, 
   java.nio.ByteOrder byteOrder,
   SerializationHandle slzHandle){
    this.HEADER_SIZE = headerSize;
    this.FOOTER_SIZE = footerSize;
    this.BYTE_ORDER = byteOrder;
    this.slzHandle = slzHandle;
  }
  /**
   * Message creation by serial id.
   * @param id Message serial id
   * @return new message instance 
   * @throws InvalidMessageException when the id supplied is invalid
   */
  public abstract Message newMessage(int id) throws InvalidMessageException;
  /**
   * Message creation by name.
   * @param name Message name
   * @return new message instance 
   * @throws InvalidMessageException when the name supplied is invalid
   */
  public abstract Message newMessage(String name)throws InvalidMessageException;

  /**
   * Abstract method that tests if this message is handled by this 
   * factory or not. 
   * This could be implemented by reflection but is not so that 
   * can be deployed more generically.
   * 
   * Overrides of this method are typically expressed for each 
   * message set using 'instanceof' for the specific message set.
   *
   * @param msg Message 
   * @throws InvalidMessageException if the message is not handled
   * by this factory.
   */
  public abstract void checkMessageType(Message msg) throws InvalidMessageException;

   /**
    * Equivalent to 'factory.serialize(msg,buf,0,buf.length)'.
    * See serialize variant with offset and length parameters.
    */
   public void serialize(Message msg, byte[] buf) throws InvalidMessageException {
     serialize(msg, buf, 0, buf.length);
   }
  /**
   * Method to serialize a message onto an array of bytes.
   * 
   * @param msg the message
   * @param buf the output buffer 
   * @param off the output buffer offset
   * @param len the usable length of the output buffer
   * (with at least HEADER_SIZE+msg.serialSize()+FOOTER_SIZE capacity)
   * @throws InvalidMessageException if the message is not handled
   * by this message factory, or is invalid according to the 
   * Message.validate() method
   * @throws BufferException if the given buffer has less than
   * the required space (this is an instance of RuntimeException)
   */
   public void serialize(Message msg, byte[] buf, int off, int len)
     throws InvalidMessageException {

     checkMessageType(msg);
     msg.validate();
  
     int msgPayload = msg.serialSize();
     int totalPayload = msgPayload + HEADER_SIZE + FOOTER_SIZE;

     if(totalPayload < len){
       throw new BufferException("Buffer is too small ("+ len +" < "+ totalPayload);
     }
     Buffer obuf = new Buffer(buf, off, len, BYTE_ORDER);
     
     if(HEADER_SIZE > 0){
       MessageHeader header = newHeader();
       header.setMessageId(msg.serialId());
       header.setPayloadSize(msgPayload);
       header.setTimestamp(slzHandle.timestamp());
       header.setSequenceId(slzHandle.nextSequenceId(msg));
       header.serialize(obuf);
     }

     msg.serialize(obuf); 

     if(FOOTER_SIZE > 0){
       MessageFooter footer = newFooter();
       footer.setChecksum(slzHandle.checksum(buf,off,HEADER_SIZE+msgPayload));
       footer.serialize(obuf);
     }
   }
   
  /**
   * Convenience deserialization method for when buffer offset is 0 
   * and usable length equals the buffer's length.
   * 'Message msg = fact.unserialize(buf)' 
   * is equivalent to 'Message msg=fact.unserialize(buf,0,buf.length);
   */
  public Message unserialize(byte[] buf) throws InvalidMessageException{
    return unserialize(buf,0,buf.length);
  }
  /**
   * Method to deserialize messages from an array of bytes.
   * 
   * @param buf the source buffer
   * @param off the source buffer offset
   * @param len the source buffer usable length
   * @throws InvalidMessageException if a message was found invalid
   * (according to validate() method or if checksum fails)
   * @return unserialized message
   */
   public Message unserialize(byte[] buf, int off, int len)
     throws InvalidMessageException {
     
     Buffer ibuf = new Buffer(buf, off, len, BYTE_ORDER);
     int msgId;
     long msgSeqId;
     double msgTimestamp;
      
     if(HEADER_SIZE > 0){
       MessageHeader header = newHeader();
       header.unserialize(ibuf);
       msgId = header.getMessageId();
       msgSeqId = header.getSequenceId();
       msgTimestamp = header.getTimestamp();
       
       if(header.getPayloadSize() != (len - HEADER_SIZE - FOOTER_SIZE)){
    	 throw new InvalidMessageException("Invalid message payload length");  
       }
     }else{
       msgId = 0;
       msgSeqId = 0L;
       msgTimestamp = slzHandle.timestamp();
     }     
     
     Message msg = newMessage(msgId);
     msg.unserialize(ibuf);
     msg.setSequenceId(msgSeqId);
     msg.setTimestamp(msgTimestamp);
     
     if(FOOTER_SIZE > 0){
       MessageFooter footer = newFooter();
       footer.unserialize(ibuf);
       if(slzHandle.checksum(buf, off, len - FOOTER_SIZE) != footer.getChecksum()){
         throw new InvalidMessageException("Checksum error : "+msg.getClass());
       }
     } 
     msg.validate(); // validate message contents
     return msg;
   }
   
   /**
    * Header instance factory method.
    * It should return null if the message set defines no headers.
    * @return new MessageHeader instance
    */
   public abstract MessageHeader newHeader();

   /**
    * Footer instance factory method.
    * It should return null if the message set defines no headers.
    * @return new MessageFooter instance
    */
   public abstract MessageFooter newFooter();

   /**
    * Convert message to byte array.
    * @param msg message to convert
    * @return byte array containing serialized message
    */
   public byte[] toByteArray(Message msg) throws InvalidMessageException {
     byte[] data = new byte[HEADER_SIZE+FOOTER_SIZE+msg.serialSize()];
     serialize(msg,data);
     return data;
   }
   
   public Message fromDataArray(byte[] data) throws InvalidMessageException {
     return unserialize(data);
   }
}
