/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;

import java.io.IOException;
import java.net.Socket;
import java.net.SocketException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketTimeoutException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * TCP client socket message link.
 *
 * <p>This message link allows simple bi-directional communication  
 * using a client TCP socket connected to a specified remote TCP server
 * socket.</p>
 * 
 * @author Eduardo Marques
 * @version 0.1
 */
public class TCPClientLink extends Link {
  /**
   * Remote server port.
   */
  private int port;
  /** 
   * Current timeout setting.
   */
  private int timeoutSetting;
  /**
   * Remote server address.
   */
  private InetAddress addr;
  /**
   * TCP socket.
   */
  private Socket sock;
  /**
   * Input stream for the TCP socket.
   */
  private InputStream inputStream;
  /**
   * Output stream for the TCP socket.
   */
  private OutputStream outputStream;

  /**
   * Constructor.
   * 
   * Note that this only initializes the TCP link, but does 
   * not start it. This is done through the connect() method
   * in the parent MessageLink class.
   *
   * @param factory the message factory to use
   * @param host host name or IP address 
   * @param port port number at remote host
   * @throws UnknownHostException as described in InetAddress.getByName()
   */
  public TCPClientLink
    (MessageFactory factory, 
     String host, 
     int port)
  throws UnknownHostException {
    super(factory);
    this.port = port;
    this.addr = InetAddress.getByName(host);
    this.timeoutSetting = 0;
    sock = null;
  }
  /**
   * Start TCP client link.
   */
  protected void start() throws IOException {
    try {
      sock = new Socket(addr,port);
      inputStream = sock.getInputStream();
      outputStream = sock.getOutputStream();
    }catch(SocketException e){
      throw new IOException(e.toString());
    }
  }
  /**
   * Stop TCP client link.
   */
  protected void stop() throws IOException {
    if(inputStream != null){
      inputStream.close();
    }
    if(outputStream != null){
      outputStream.close();
    }
    if(sock != null){
      sock.close();
      sock = null;
    }
  }
  /**
   * Receive a message packet.
   *  
   * @param buf read buffer
   * @param off read buffer offset
   * @param len read buffer usable length
   * @param timeout timeout for receive
   * @return the next datagram packet's payload length or 0 
   * if timeout has been reached
   * @throws IOException if a network error occurs
   */
  protected int recv(byte[] buf, int off, int len, int timeout) 
  throws IOException {
    if(timeoutSetting != timeout){
      try { 
        sock.setSoTimeout(timeout); 
        timeoutSetting = timeout;
      } 
      catch(SocketException e){
        throw new IOException(e.toString());
      }
    }
    try {
      len = inputStream.read(buf, off, len);
    }catch(SocketTimeoutException e){
      return 0; // timeout
    }
    return len;
  }
  /**
   * Send a message packet.
   * 
   * @throws IOException if a network error occurs
   */
  protected void send(byte[] data, int off, int len) throws IOException {
    outputStream.write(data, off, len);
    outputStream.flush();
  }
}
