/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;

/**
 * TCP server socket message link.
 *
 * <p>This message link allows simple bi-directional communication  
 * using a server TCP socket.</p>
 *
 * <p>Note that the TCP server link may have multiple peers.
 * Messages from all peers will be placed on the link's receive port
 * and any messages in the link's send port will be sent to all
 * connected peers at any time.</p>
 * 
 * @author Eduardo Marques
 * @version 0.1
 */
public class TCPServerLink extends Link {
  /**
   * Listening port.
   */
  private int port;
  
  /**
   * Server channel for "select" operation.
   */
  private ServerSocketChannel serverCh;
  /**
   * Selector for "select" operation.
   */
  private Selector selector;

  /**
   * List of client SocketChannel handles.
   */
  private ArrayList clients;
  
  /**
   * Constructor.
   * 
   * Note that this only initializes the TCP link, but does 
   * not connect it. This is done through the start() method
   * in the parent MessageLink class.
   *
   * @param factory the message factory to use
   * @param port server port to use
   */
  public TCPServerLink(MessageFactory factory, int port){
    super(factory);
    this.clients = new ArrayList();
    this.port = port;
    serverCh = null;
  }
  
  /**
   * Start the TCP server link.
   * <br>This will open and setup the link's TCP server socket.
   * @throws IOException if a network error occurs
   */
  protected void start() throws IOException {
    try {
      selector = Selector.open();
      serverCh = ServerSocketChannel.open();
      serverCh.configureBlocking(false);
      serverCh.socket().bind(new InetSocketAddress(port));
      serverCh.register(selector, SelectionKey.OP_ACCEPT);
    }catch(SocketException e){
      throw new IOException(e.toString());
    }
  }
  
  /**
   * Stop the TCP server link.
   * <br>This will disconnect the link's TCP server socket.
   * @throws IOException if a network error occurs
   */
  protected void stop() throws IOException {
    selector.close();
    for(int i=0;i<clients.size();++i){
      SocketChannel channel = (SocketChannel) clients.get(i);
      try {
       channel.close();
      }catch(IOException e){
        System.err.println("Error closing client socket: "+e.toString());
      }
    }
    if(serverCh != null){
      serverCh.close();
      serverCh = null;
    }
  }
  
  /**
   * Receive data over the TCP server link.
   * <br>Note that multiple messages (from potential multiple clients)
   * will be placed onto the output buffer.
   * @param buf read buffer
   * @param off read buffer offset
   * @param len read buffer usable length
   * @param timeout timeout for receive
   * @return the read data length or 0 if timeout has been reached
   * @throws IOException if a network error occurs
   */
  protected int recv(byte[] buf, int off, int len, int timeout) 
  throws IOException {
    if(selector.select(timeout) == 0) return 0;
    int readData=0;

    Set keys = selector.selectedKeys();
    Iterator it = keys.iterator();
    
    while(it.hasNext()){
      SelectionKey key=(SelectionKey) it.next();
      it.remove();  
      if(!key.isValid()){
        System.err.println("ERR "+key.channel());
        if(key.channel() instanceof SocketChannel){
          key.channel().close();
          clients.remove(key.channel());
        }
      }
      else 
      if((key.readyOps() & SelectionKey.OP_ACCEPT) == SelectionKey.OP_ACCEPT) {
        // New client connection
        SocketChannel channel = serverCh.accept();
        if(channel == null) continue;
        channel.configureBlocking(false);
        channel.register(selector, SelectionKey.OP_READ);
        synchronized(clients){
          clients.add(channel);
        }
        //System.err.println("New client: "+channel);
      }
      else 
      if((key.readyOps() & SelectionKey.OP_READ) == SelectionKey.OP_READ){
        // Data available
        SocketChannel channel = (SocketChannel) key.channel();    
        if(readData < len){
          try {
            readData += channel.read(ByteBuffer.wrap(buf,off+readData,len-readData));
          }catch(IOException e){
            synchronized(clients){ clients.remove(channel); }
            System.err.println("Lost touch with "+channel + "??? -> "+e.getMessage());
            channel.close();
          }
        }
      }
    }
    return readData;
  }
  /**
   * Send data over the TCP server link.
   * <br>Note that all active client peers get the data sent.
   * @throws IOException if a network error occurs
   */
  protected void send(byte[] data, int off, int len) throws IOException {
    Iterator it = clients.iterator();
    while(it.hasNext()){
      SocketChannel channel = (SocketChannel) it.next();
      try{ 
        channel.write(ByteBuffer.wrap(data, off, len));
      }catch(IOException e){
        it.remove();
        System.out.println("Lost touch with "+channel + "??? -> "+e.getMessage());
        channel.close();
      }
    }
  }
}
