/*****************************************************************************/
/* This code is part of the Jarol project: jarol.cs.uni-salzburg.at          */
/*                                                                           */
/* Copyright (c) 2006-2007 Eduardo Marques                                   */
/*                                                                           */
/* This program is free software; you can redistribute it and/or modify      */
/* it under the terms of the GNU General Public License as published by      */
/* the Free Software Foundation; either version 2 of the License, or         */
/* (at your option) any later version.                                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the              */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software Foundation,   */
/* Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.        */
/*                                                                           */
/*****************************************************************************/
package jarol.messages;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.SocketTimeoutException;


/**
 * UDP socket message link.
 *
 * <p>This message link allows simple bi-directional communication  
 * between UDP ports. The local host will listen on a specified
 * local port for messages and send messages to a specified UDP
 * port on a remote host.</p>
 * 
 * @author Eduardo Marques
 * @version 0.1
 */
public class UDPLink extends Link {
  /**
   * UDP datagram socket.
   */
  private DatagramSocket sock;

  /**
   * Local UDP port for receiving data.
   */
  private int localPort;
  
  /**
   * Remote UDP port to send data to.
   */
  private int remotePort;
  
  /**
   * Current receive timeout setting.
   */
  private int timeoutSetting;
  
  /**
   * Remote UDP socket address.
   */
  private InetAddress addr;

  /**
   * Constructor.
   * 
   * Note that this only initializes the UDP link, but does 
   * not start it. This is done through the connect() method
   * in the parent MessageLink class.
   *
   * @param factory the message factory to use
   * @param host host name or IP address 
   * @param localPort local UDP port 
   * @param remotePort remote UDP port
   * @throws UnknownHostException as described in InetAddress.getByName()
   */
  public UDPLink
    (MessageFactory factory, 
     String host, 
     int localPort, 
     int remotePort)
  throws UnknownHostException {
    super(factory);
    this.localPort = localPort;
    this.remotePort = remotePort;
    this.addr = InetAddress.getByName(host);
    this.timeoutSetting = 0;
    sock = null;
  }
  /**
   * Constructor with specified receive / send buffer size.
   * 
   * Note that this only initializes the UDP link, but does 
   * not start it. This is done through the connect() method
   * in the parent MessageLink class.
   *
   * @param factory the message factory to use
   * @param host host name or IP address 
   * @param localPort local UDP port 
   * @param remotePort remote UDP port
   * @param recvBufSz receive buffer size
   * @param sendBufSz send buffer size
   * @throws UnknownHostException as described in InetAddress.getByName()
   */
  public UDPLink
    (MessageFactory factory, 
     String host, 
     int localPort, 
     int remotePort,
     int recvBufSz,
     int sendBufSz)
  throws UnknownHostException {
    super(factory,recvBufSz,sendBufSz);
    this.localPort = localPort;
    this.remotePort = remotePort;
    this.addr = InetAddress.getByName(host);
    this.timeoutSetting = 0;
    sock = null;
  }
  
  /**
   * Start the UDP link.
   * <p>This will set up the UDP datagram socket listening onto
   * the local UDP port.</p>
   * @throws IOException if an error occurs creating the socket.
   */
  protected void start() throws IOException {
    try {
      sock = new DatagramSocket(localPort);
    }catch(SocketException e){
      throw new IOException(e.toString());
    }
  }
  /**
   * Stop the UDP link.
   * <p>This will close the UDP datagram socket.</p>
   * @throws IOException if an error occurs closing the socket.
   */
  protected void stop() throws IOException {
    if(sock != null){
      sock.close();
      sock = null;
    }
  }
  /**
   * Receive data.
   *  
   * @param buf read buffer
   * @param off read buffer offset
   * @param len read buffer usable length
   * @param timeout timeout for receive
   * @return the next datagram packet's payload length or 0 
   * if timeout has been reached
   * @throws IOException if a network error occurs
   */
  protected int recv(byte[] buf, int off, int len, int timeout) 
  throws IOException {
    DatagramPacket packet = new DatagramPacket(buf, off, len);
    if(timeoutSetting != timeout){
      try { 
        sock.setSoTimeout(timeout); 
        timeoutSetting = timeout;
      } 
      catch(SocketException e){
        throw new IOException(e.toString());
      }
    }
    try {
      sock.receive(packet);
    }catch(SocketTimeoutException e){
      return 0; // timeout
    }
    return packet.getLength();
  }

  /**
   * Send data to the remote UDP socket.
   *
   * @param data data buffer
   * @param off data buffer offset
   * @param len data buffer usable length
   * @throws IOException if a network error occurs
   */
  protected void send(byte[] data, int off, int len) throws IOException {
    sock.send(new DatagramPacket(data, off, len, addr, remotePort));
  }
}
